class tagPattern:
    def match(self, word, index=0):
        if index >= len(word):
            return None
        if word[index] != '<':
            return None
        index = index + 1
        if index >= len(word):
            return None
        while word[index] != '>':
            index = index + 1
            if index >= len(word):
                return None
        return index + 1

class atagPattern:
    def match(self, word, index=0):
        if index >= len(word):
            return None
        if word[index:index+2].lower() != '<a':
            return None
        index = index + 2
        if index >= len(word):
            return None
        while word[index:index+3].lower() != '</a':
            index = index + 1
            if index + 2 >= len(word):
                return None
        index = index + 3
        while word[index] != '>':
            index = index + 1
            if index >= len(word):
                return None
        return index + 1

tag_pattern = tagPattern()
atag_pattern = atagPattern()

lowerAlphabets = 'abcdefghijklmnopqrstuvwxyz'
upperAlphabets = lowerAlphabets.upper()
allAlphabets = lowerAlphabets + upperAlphabets
minAlphabet = min(allAlphabets)
maxAlphabet = max(allAlphabets)

def is_alphabet(char):
    return min(char, minAlphabet) == minAlphabet and max(char, maxAlphabet) == maxAlphabet

class wordTree:

    def __init__(self, dict={}):
	self.children=dict.copy()

    def add(self, word):
        if len(word)>0:
            if self.children.has_key(word[0]):
                self.children[word[0]].add(word[1:])
            else:
                child = wordTree({None:None})
                index = len(word)-1
                while index > 0:
                    child = wordTree({word[index]:child})
                    index = index - 1
                self.children[word[0]]=child
        else:
            self.children[None]=None

    def match(self, string, index=0, separated=None):
        if separated and index > 0 and is_alphabet(string[index-1]):
            return None
        if index > len(string):
            return None
        end = len(string)
        node = self
        if node.children.has_key(None):
            lastMatchPosition = index
        else:
            lastMatchPosition = None
        while index < end:
            if node.children.has_key(string[index]):
                node = node.children[string[index]]
                index = index + 1
                if node.children.has_key(None):
                    if not(separated and index < end and is_alphabet(string[index])):
                        lastMatchPosition = index
            else:
                return lastMatchPosition
        return lastMatchPosition

    def find_first_match(self, string, index=0, separated=None):
        end = len(string)
        while index <= end:
            matched = self.match(string, index, separated)
            if matched:
                return index, matched
            index = index + 1
        return None

    def find_matches(self, string, index=0, separated=None):
        end = len(string)
        matches = []
        while index <= end:
            matchedPosition=self.match(string, index, separated)
            if matchedPosition is not None:
                matches = matches+[(index, matchedPosition)]
                index = matchedPosition + 1
            else:
                index = index + 1
        return matches

class htmlWordTree(wordTree):
    def find_matches(self, string, index=0, separated=None):
        end = len(string)
        matches = []
        while index <= end:
            skip = tag_pattern.match(string, index)
            if skip:
                skip2 = atag_pattern.match(string, index)
                if skip2:
                    index = skip2
                else:
                    index = skip
            else:
                matchedPosition=self.match(string, index, separated)
                if matchedPosition is not None:
                    matches = matches+[(index, matchedPosition)]
                    index = matchedPosition + 1
                else:
                    index = index + 1
        return matches
        
    def find_first_match(self, string, index=0, separated=None):
        end = len(string)
        while index <= end:
            skip = tag_pattern.match(string, index, separated)
            if skip:
                skip2 = atag_pattern.match(string, index, separated)
                if skip2:
                    index = skip2
                else:
                    index = skip
            else:
                matched = self.match(string, index, separated)
                if matched:
                    return index, matched
                index = index + 1
        return None

maxSmallNum = 1000

class Histogram:
    def __init__(self, dict = {}):
        self._words = dict.copy()
        self._cached_total = None
        self._cached_max = None

    def add(self, string):
        self._cached_total = None
        self._cached_max = None
        if self._words.has_key(string):
            value = self._words[string]
            carry = 1
            index = 0
            while index < len(value) and carry != 0:
                digit = value[index]+carry
                value[index] = digit % maxSmallNum
                carry = digit / maxSmallNum
                index = index + 1
            if carry != 0:
                value.append(carry)
        else:
            self._words[string] = [1]

    def percent(self, string):
        if not(self._words.has_key(string)):
            return 0
        value = self._words[string]
        valueLen = len(value)
        total = self._total()
        totalLen = len(total)
        if totalLen == 1:
            return 100 * value[-1] / total[-1]
        totalNum = total[-1] * maxSmallNum + total[-2]
        if totalLen == valueLen:
            valueNum = value[-1] * maxSmallNum + value[-2]
        else:
            if totalLen - valueLen == 1:
                valueNum = value[-1]
            else:
                valueNum = 0
        return 100 * valueNum / totalNum

    def words(self):
        words = []
        for w in self._words.keys():
            words.append(w)
        return words

    def percentToMax(self, string):
        if not(self._words.has_key(string)):
            return 0
        value = self._words[string]
        valueLen = len(value)
        total = self._max()
        totalLen = len(total)
        if totalLen == 1:
            return 100 * value[-1] / total[-1]
        totalNum = total[-1] * maxSmallNum + total[-2]
        if totalLen == valueLen:
            valueNum = value[-1] * maxSmallNum + value[-2]
        else:
            if totalLen - valueLen == 1:
                valueNum = value[-1]
            else:
                valueNum = 0
        return 100 * valueNum / totalNum


    def _total(self):
        if self._cached_total:
            return self._cached_total
        keys = self._words.keys()
        maxLen = 0
        for key in keys:
            maxLen = max(maxLen, len(self._words[key]))
        total = []
        carry = 0
        for index in range(maxLen):
            sum = carry
            for key in keys:
                if len(self._words[key]) > index:
                    sum = sum + self._words[key][index]
            total.append(sum % maxSmallNum)
            carry = carry / maxSmallNum
        while carry != 0:
            total.append(carry % maxSmallNum)
            carry = carry / maxSmallNum
        self._cached_total = total
        return total

    def _max(self):
        if self._cached_max:
            return self._cached_max
        keys = self._words.keys()
        maxLen = 0
        for key in keys:
            maxLen = max(maxLen, len(self._words[key]))
        maxValue = []
        for i in range(maxLen):
            maxValue.append(0)
        for key in keys:
            if len(self._words[key]) == maxLen:
                value = self._words[key]
                index = maxLen - 1
                while index >= 0:
                    if value[index] > maxValue[index]:
                        maxValue = value
                        break
                    else:
                        if value[index] < maxValue[index]:
                            break
                    index = index - 1
        self._cached_max = maxValue
        return maxValue
