__version__ = "0.1"

__all__ = ["lomScope"]

import string
import socket
import SocketServer
import urllib
import re
import lomInterface
import glossary
import UserModel

"""
constants for referent LivingOM... should be specifiable by user!
"""
host = 'webguide.cs.colorado.edu'
#host = 'localhost'
port = 9080
livingOM = 'livingOM'
#livingOM = 'entwine'
url = 'http://%s:%d/%s/Concepts/manage_autoLinkForm'%(host, port, livingOM)
request_key_pattern = '?url=%s&text=&file='
lomUrl = 'http://%s:%d/%s/Concepts/'%(host, port, livingOM)

"""
constants for proxy control buttons
"""
on_color = '#80f080'
off_color = '#406040'
always_color = '#f0f080'
body_pattern = re.compile("(?i)(.*<body[^>]*>)(.*)")
reloadConcepts_url = 'http://lomScope:0/reloadConcepts?url=%s'
reloadConcepts_pattern = re.compile('(?i)http://lomScope:0/reloadConcepts\?url=(.*)')
def reloadConcepts_td(target_url):
    return '<td bgcolor=%s><a href="%s">Reload Concepts</a></td>'%(always_color,reloadConcepts_url%(urllib.quote(target_url)))

originalMode_url = 'http://lomScope:0/originalMode?url=%s'
originalMode_pattern = re.compile('(?i)http://lomScope:0/originalMode\?url=(.*)')

autoLinkMode_url = 'http://lomScope:0/autoLinkMode?url=%s'
autoLinkMode_pattern = re.compile('(?i)http://lomScope:0/autoLinkMode\?url=(.*)')
def autoLink_td(target_url):
    if config.autoLink:
        url = originalMode_url%(urllib.quote(target_url))
        color = on_color
    else:
        url = autoLinkMode_url%(urllib.quote(target_url))
        color = off_color
    return '<td bgcolor=%s><a href="%s">AutoLink</a></td>'%(color, url)

separatedMode_url = 'http://lomScope:0/separatedMode?url=%s'
separatedMode_pattern = re.compile('(?i)http://lomScope:0/separatedMode\?url=(.*)')
nonSeparatedMode_url = 'http://lomScope:0/nonSeparatedMode?url=%s'
nonSeparatedMode_pattern = re.compile('(?i)http://lomScope:0/nonSeparatedMode\?url=(.*)')
def separatedMode_td(target_url):
    if config.separated:
        url = nonSeparatedMode_url%(urllib.quote(target_url))
        color = on_color
    else:
        url = separatedMode_url%(urllib.quote(target_url))
        color = off_color
    return '<td bgcolor=%s><a href="%s">Word Sep</a></td>'%(color, url)


"""
HTTP utilities
"""
keep_alive_pattern = re.compile("(?i)(.*\\r\\n)keep-alive:[^\\r]*\\r\\n(.*)")
def  forward_response(url):
    return 'HTTP/1.0 302 Found\r\nLocation: %s\r\nConnection: close\r\nContent-Type: text/html; charset=iso-8859-1\r\n\r\n<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">\r\n<HTML><HEAD>\r\n<TITLE>302 Found</TITLE>\r\n</HEAD><BODY>\r\n<H1>Found</H1>\r\nClick here to go back to <A HREF="%s">the document</A>.<P>\r\n<HR>\r\n</BODY></HTML>\r\n'%(url,url)
def html_response(string):
    return 'HTTP/1.0 200 OK\r\nConnection: close\r\nContent-Type: text/html\r\n\r\nContent-Length: %d\r\n\r\n%s'%(len(string), string)

"""
utility object for auto-linking to LivingOM's concepts
"""
lomAutoLinker = lomInterface.lomAutoLinker(lomUrl)

"""
configuration manager
"""
class lomScopeConfig:
    def __init__(self):
        self.autoLink = 1
        self.separated = 1

config = lomScopeConfig()

"""
user model
"""
userModel = UserModel.userModel()

class lomScope(SocketServer.StreamRequestHandler):
    def handle(self):
        request = self.rfile.readline()
        while 1:
            requestLine = self.rfile.readline()
            if requestLine == '\r\n':
              break
            if requestLine == '\n':
              break
            request = request + requestLine
        self.process_request(request)

    def process_request(self,request):
        target_url = string.split(string.split(request, '\r\n')[0], ' ')[1]
        # reloading concepts from LivingOM
        if reloadConcepts_pattern.match(target_url):
            lomAutoLinker.readConcepts()
            quoted_url,dummy = reloadConcepts_pattern.subn('\\1', target_url)
            self.wfile.write(forward_response(urllib.unquote(quoted_url)))
            return
        # turning all modes off
        if originalMode_pattern.match(target_url):
            config.autoLink = None
            quoted_url,dummy = originalMode_pattern.subn('\\1', target_url)
            self.wfile.write(forward_response(urllib.unquote(quoted_url)))
            return
        # turning autoLink mode on
        if autoLinkMode_pattern.match(target_url):
            config.autoLink = 1
            quoted_url,dummy = autoLinkMode_pattern.subn('\\1', target_url)
            self.wfile.write(forward_response(urllib.unquote(quoted_url)))
            return
        if separatedMode_pattern.match(target_url):
            config.separated = 1
            quoted_url,dummy = nonSeparatedMode_pattern.subn('\\1', target_url)
            self.wfile.write(forward_response(urllib.unquote(quoted_url)))
            return
        if nonSeparatedMode_pattern.match(target_url):
            config.separated = None
            quoted_url,dummy = nonSeparatedMode_pattern.subn('\\1', target_url)
            self.wfile.write(forward_response(urllib.unquote(quoted_url)))
            return
        # do proxy
        self.do_proxy(request)

    def do_proxy(self, request):

        do_autoLink = config.autoLink
        request, n = keep_alive_pattern.subn('\\1keep-alive: close\r\n\\2', request)
	if n == 0:
	    request = request + 'keep-alive: close\r\n'
        command = string.split(request, '\r\n')[0]+'\r\n'
        command_words = string.split(command, ' ')
        if do_autoLink and command_words[0] != 'GET':
            do_autoLink = None
        target_url = command_words[1]
        isHtml = self.isHtml(target_url)
        if do_autoLink and not(isHtml):
            do_autoLink = None
        target_type, target_opaque = urllib.splittype(target_url)
        target_host, target_path = urllib.splithost(target_opaque)
        target_host, target_port = urllib.splitport(target_host)
        if target_port is None:
          target_port = 80
        else:
          target_port = string.atoi(target_port)
        add, dummy = self.client_address
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((target_host, target_port))
        s.send('%s %s HTTP/1.0\r\n'%(command_words[0],command_words[1]))
        print '%s %s HTTP/1.0\r\n'%(command_words[0],command_words[1])
        s.send(request[len(command):])
        s.send('\r\n')
        f = s.makefile()
        content_length=0
        while 1:
            m=f.readline()
            if string.lower(m[:15]) == 'content-length:':
                content_length=string.atoi(m[16:])
            else:
                if m == '' or m == '\r\n' or m == '\r\n\r\n':
                    break
                else:
                    self.wfile.write(m)
        if content_length > 0:
            m=f.read(content_length)
        else:
	    m=f.read()
        if isHtml:
            userModel.incorporate(m, lomAutoLinker.wordTree,separated=config.separated)
            userModel.save()
        if do_autoLink:
            print 'auto linking...'
            m=lomAutoLinker.autoLink(m, color='#ffff80', words = userModel.wordsLessThan(50),separated=config.separated)
            m=lomAutoLinker.autoLink(m, color='#ff8080', words = userModel.wordsMoreThan(50),separated=config.separated)
	    print 'done'
        if content_length > 0:
	    self.wfile.write('content-length: %d\r\n\r\n'%(len(m)))
	else:
	    self.wfile.write('\r\n')
        if isHtml:
            m=self.add_control(m,target_url)
        self.wfile.write(m)
        f.close
        s.close

    def add_control(self, string, target_url):
        tr = autoLink_td(target_url)
        tr = tr + separatedMode_td(target_url)
        tr = tr + '<td></td>'+ reloadConcepts_td(target_url)
        string, dummy = body_pattern.subn('\\1<p><table border="3" color=black><tr bgcolor=#404040>'+tr+'</tr></table></p><hr>\\2', string)
        return string

    def isHtml(self, target_url):
        target_type, target_opaque = urllib.splittype(target_url)
        target_host, target_path = urllib.splithost(target_opaque)
        target_host, target_port = urllib.splitport(target_host)
        if target_port is None:
          target_port = 80
        else:
          target_port = string.atoi(target_port)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((target_host, target_port))
        f = s.makefile()
        s.send('HEAD %s HTTP/1.0\r\n\r\n'%(target_url))
        content_type = None
        while 1:
          response = f.readline()
          if string.lower(response[:13]) == 'content-type:':
            content_type = response[14:]
            while content_type[-1:] == '\n': content_type = content_type[:-1]
            while content_type[-1:] == '\r': content_type = content_type[:-1]
            break
        f.close
        s.close
        return content_type == 'text/html'

def test(HandlerClass = lomScope,
         ServerClass = SocketServer.TCPServer):
    """Test the HTTP request handler class.

    This runs an HTTP server on port 8000 (or the first command line
    argument).

    """

    import sys

    if sys.argv[1:]:
        port = int(sys.argv[1])
    else:
        port = 8000
    server_address = ('', port)

    httpd = ServerClass(server_address, HandlerClass)

    sa = httpd.socket.getsockname()
    print "Serving HTTP on", sa[0], "port", sa[1], "..."
    httpd.serve_forever()


if __name__ == '__main__':
    test()
